--- ### DUNGEONS

local A = select(2, ...);
local C = A.Constants;
local U = A.Utils;
local API = A.Api.Dungeon;
local ShortsMixin = A.Mixins.Shorts;
local ShortsApiMixin = A.Mixins.ShortsApi;
local ModeApi = A.Api.Mode;

-------- STORES -----------------

local shortsStore = {};
local dungeons = {};
local dungeonsById = {};
local dungeonsByNameMode = {};


-------- DUNGEON ITEM -----------------

local ItemMixin = {}
function ItemMixin:_getShortsStore()
    return shortsStore
end

function ItemMixin:IsHeroic()
    return ModeApi:IsHeroic(self.mode);
end

function ItemMixin:IsRaid()
     return self.groupType ~= C.DungeonSizes.Group;
end

function ItemMixin:ToString()
    if self:IsHeroic() then
        return self.name .. " (" .. self.mode .. ")";
    else
        return self.name;
    end
end

local colors = {
    default = { 1,1,1 },

    byMode = {
        [true] = { 1, .5, 0 },
        [false] = { .6, 1, .5 }
    },

    byRaid = {
        [true] = { ITEM_QUALITY_COLORS[4].r, ITEM_QUALITY_COLORS[4].g, ITEM_QUALITY_COLORS[4].b  } --{ .6, .5, 1 }
    },

    GetColor = function(self, dungeon, asHex)
        local clr = self.default;

        if dungeon:IsRaid() then
            clr = self.byRaid[true];
        else
            clr = self.byMode[dungeon:IsHeroic()];
        end
        if asHex then
            return U:RgbToHex(clr);
        else
            return clr;
        end
    end
}

function ItemMixin:GetColor(asHex)
    return colors:GetColor(self, asHex);
end

-------- COMMON API -----------------

Mixin(API, ShortsApiMixin);
function API:_getShortsStore()
    return shortsStore
end
function API:GetById(id)
    return dungeonsById[id];
end

-- AddShorts
-- GetByShort




function API:IsMatch(dungeon, filters)

    if filters.era ~= nil and dungeon.era ~= filters.era then
        return false;
    end


    if filters.heroic ~= nil and dungeon:IsHeroic() ~= filters.heroic then
        return false;
    end

    if filters.name ~= nil
        and string.find(dungeon.normalizedName, filters.name) == nil 
        and string.find(dungeon.engNormalizedName, filters.name) == nil
        then
        return false;
    end

    if filters.raid ~= nil and dungeon:IsRaid() ~= filters.raid then
        return false;
    end

    return true;
end

function API:GetAll(filters)
    return self:FilterTable(dungeons, filters);
end

function API:GetByName(name, mode)
    if name == nil then return end
    name = string.lower(name);
    if mode == nil then
        mode = ModeApi.Normal.id;
    end
    return dungeonsByNameMode[name .. C.RequestModeSeparator .. mode];
end

function API:GetOne(id, arg)
    if arg == nil then
        return self:GetById(id);
    else
        return self:GetByName(id, arg);
    end
end


function API:FilterTable(items, filters)
    if filters == nil then
        return items;
    else
        local res = {};
        for _, item in pairs(items) do
            if self:IsMatch(item, filters) then
                table.insert(res, item);
            end
        end
        return res;
    end
end

function API:TryGues(arg)
    local d = self:GetById(arg);
    if d ~= nil then return d end

    local name,mode = unpack(U:TextSplit(arg, C.RequestModeSeparator));

    d = self:GetByName(name, mode);
    if d ~= nil then return d end

    return self:FindByShort(arg);
end

function API:SwapMode(dung)
    local name = dung.normalizedName;
    if dung.mode == ModeApi.Normal.id then
        return self:GetOne(name, ModeApi.Heroic.id) or dung;
    else
        return self:GetOne(name, ModeApi.Normal.id) or dung;
    end
end

function API:_inspect(what)
    if what == "shortsStore" then return shortsStore end
    if what == "dungeons"  then return dungeons end
    if what == "dungeonsById" then return dungeonsById end
    if what == "dungeonsByNameMode" then return dungeonsByNameMode end
end

-------- INITIALIZE -----------------


local function CreateDungeon(id, name, era, groupType, mode)
    local d = {
        id = id .. C.RequestModeSeparator .. mode.id,
        key = id,
        name = name,
        normalizedName = string.lower(name),
        engNormalizedName = string.lower(id),
        era = era,
        groupType = groupType,
        mode = mode.id,
        modeName = mode.name,
        enabled = false,
        shorts = { }
    }
    Mixin(d, ItemMixin, ShortsMixin);
    
    local shorts = C.Shorts.Dungeons[id] or {}
    
    d:AddShorts(id, name);
    d:AddShorts(unpack(shorts));

    return d;
end

local function InsertDungeon(dungeon)
    table.insert(dungeons, dungeon);
    dungeonsById[dungeon.id] = dungeon;
    dungeonsByNameMode[dungeon.normalizedName .. C.RequestModeSeparator .. dungeon.mode] = dungeon;
end

local process_array = {
    -- { array, era, has heroic, dungeon_size }
    { C.CL_Dungeons, C.Eras.Classic, false, C.DungeonSizes.Group },
    { C.CL_Raids10, C.Eras.Classic, false, C.DungeonSizes.Raid10 },
    { C.CL_Raids20, C.Eras.Classic, false, C.DungeonSizes.Raid20 },
    { C.CL_Raids40, C.Eras.Classic, false, C.DungeonSizes.Raid40 },
    { C.BC_Dungeons, C.Eras.BC, true, C.DungeonSizes.Group },
    { C.BC_Raids10, C.Eras.BC, false, C.DungeonSizes.Raid10 },
    { C.BC_Raids25, C.Eras.BC, false, C.DungeonSizes.Raid25 },
}

for _, _arr in pairs(process_array) do
    local arr, era, hasHeroic, groupTypeId = unpack(_arr);
    for id, name in pairs(arr) do
        InsertDungeon(CreateDungeon(id, name, era, groupTypeId, ModeApi.Normal));
        if hasHeroic then
            InsertDungeon(CreateDungeon(id, name, era, groupTypeId, ModeApi.Heroic));
        end
    end
end

table.sort(dungeons, function(a,b) 
    if a.name == b.name then
        local ah = a:IsHeroic() and 1 or 0;
        local bh = b:IsHeroic() and 1 or 0;
        return ah < bh;
    else
        return a.name < b.name;
    end
end)




