local MAJOR, MINOR = "TabuFrameBuilder-2.0", 2 -- 8.2.0 v2 / increase manually on changes
local fb = LibStub:NewLibrary(MAJOR, MINOR);
if not fb then 
	return;
end




local function COALESCE(...)
	for i=1, select("#",...) do
		local value = select(i, ...);
		if value ~= nil then return value end
	end
end
function fb:EnsureObject(obj)
    if obj == nil then return {} end
    return obj;
end
------------------------------

function fb:ApplyFrameBuilders(arg, builders)
	local frame;

	if type(arg) == "function" then
		frame = arg();
	else
		frame = arg;
	end

	local callback_args = { };
	for _i, callback in pairs(builders) do
		callback_args = { callback(frame, unpack(callback_args)) }
	end

	return frame;
end


------------------------------

local SetFrameDimension = function(frame, value, dimension)
	if value ~= nil and value >= 0 then
		dimension(frame, value);
		return true;
	end
	return false;
end

function fb:SetSize(frame, size)
	if not size then return false end
    local w = self:SetWidth(frame, size[1]);
    local h = self:SetHeight(frame, size[2]);
	return w and h;
end

function fb:SetWidth(frame, value)
	return SetFrameDimension(frame, value, frame.SetWidth);
end

function fb:SetHeight(frame, value)
	return SetFrameDimension(frame, value, frame.SetHeight);
end

function fb:SetupSize(frame, options)
	options = self:EnsureObject(options);

	if self:SetSize(frame, options.size) then
		return
	end
	self:SetWidth(frame, options.width);
	self:SetHeight(frame, options.height);
end

---------------------------

function fb:SetPoint(frame, pointOptions, parentFrame, shouldNotClear)

	if not pointOptions then return end

	local setpoint = { unpack(pointOptions) };

    if type(setpoint[3]) == "string" then
        if setpoint[2] == "parent" then
            setpoint[2] = parentFrame;
        end
    end

	if not shouldNotClear then
		frame:ClearAllPoints();
	end

	frame:SetPoint(unpack(setpoint));

end

function fb:SetPoints(frame, pointsOptions, parentFrame, shouldNotClear)

    if not shouldNotClear then
        frame:ClearAllPoints();
    end

    local shouldNotClearPoint = true;
    for _i, pointOptions in pairs(pointsOptions) do
        self:SetPoint(frame, pointOptions, parentFrame, shouldNotClearPoint);
    end

end

function fb:SetupAnchor(frame, options, parentFrame)
	if not options then return end
	if options.points == true then
		frame:ClearAllPoints();
		frame:SetAllPoints();
	elseif options.point then
		self:SetPoint(frame, options.point, parentFrame);
	elseif options.points then
        self:SetPoints(frame, options.points, parentFrame);
	end
end

------------------------------

function fb:SetChild(parent, key, child)
	if parent and key then
		parent[key] = child;
	end
end

function fb:SetupChild(parent, options, child)
    if not options or not options.parentKey then return end
    self:SetChild(parent, options.parentKey, child);
end

------------------------------

function fb:SetupTextures(frame, options, parentFrame)
    if not options then return end

    if options.textures then
        for _i, texture in pairs(options.textures) do
            self:SetTexture(frame, texture, parentFrame);    
        end
    elseif options.texture then
        self:SetTexture(frame, options.texture, parentFrame);
    end
end

function fb:SetTexture(frame, options, parentFrame)
    if not options then return end
    local layer = options.layer or "OVERLAY";
    local tex = frame:CreateTexture(options.globalName, layer, options.inheritsFrom);
    tex:SetTexture(options.texture);
    if options.texCoord then
        tex:SetTexCoord(unpack(options.texCoord));
    end
    self:SetupChild(frame, options, tex);
    self:SetupSize(tex, options);
    self:SetupAnchor(tex, options, frame);
    if options.hidden then
        tex:Hide();
    end
    -- Tabu.print(options)
end

--------------------------------

function fb:SetupFontStrings(frame, options, parentFrame)
    if not options then return end;
    fb:SetFontStrings(frame, options.fontstrings, parentFrame);
end

function fb:SetFontStrings(frame, fontstrings, parentFrame)
    if not fontstrings then return end
    for _i, options in pairs(fontstrings) do
        local fs = frame:CreateFontString(options.globalName, options.layer, options.inheritsFrom);
        self:SetupAnchor(fs, options, frame);
        self:SetupSize(fs, options, frame);
        self:SetupChild(frame, options, fs);
        if options.text then
            fs:SetText(options.text);
        end
    end
end

--------------------------------

function fb:SetupScripts(frame, options, parentFrame)
    if not options then return end
    fb:SetScripts(frame, options.scripts, parentFrame);
end

function fb:SetScripts(frame, scripts, parentFrame)
    if type(scripts) ~= "table" then return end

    for key, func in pairs(scripts) do
        if type(func) == "function" then
            frame:SetScript(key, func);
        end
    end

end

--------------------------------


function fb:SetupFrame(frame, options, parentFrame)

    self:SetupSize(frame, options, parentFrame);
    self:SetupAnchor(frame, options, parentFrame);
    self:SetupChild(parentFrame, options, frame);
    self:SetupTextures(frame, options, parentFrame);
    self:SetupFontStrings(frame, options, parentFrame);
    self:SetupScripts(frame, options, parentFrame);

end

--------------------------------


function fb:MakeMovable(frame, triggerFrame, afterStart, afterStop)
	frame:EnableMouse(true);
	frame:SetMovable(true);
	
	triggerFrame = triggerFrame or frame;

	triggerFrame.movableFrame = frame;
	triggerFrame:RegisterForDrag("LeftButton")
	triggerFrame:SetScript("OnDragStart", function(self) 

		--local mframe = self.movableFrame;
		if not frame.isLocked then
			frame:StartMoving()
		end
		if type(afterStart) == "function" then
			afterStart(frame, triggerFrame);
		end
	end)
	
	triggerFrame:SetScript("OnDragStop", function(self) 
		--local mframe = self.movableFrame;
		frame:StopMovingOrSizing();
		if type(afterStop) == "function" then
			afterStop(frame, triggerFrame);
		end
	end)
	
end

-----------------------

function fb:AddScript(frame, scriptName, callback, prepend)
    local oldScript = frame:GetScript(scriptName);
    
    local newScript = function(...) 
        if prepend then
            callback(...)
        end
        if oldScript then
            oldScript(...)
        end
        if not prepend then
            callback(...)
        end
    end
    frame:SetScript(scriptName, newScript);
end



-----------------------

function fb:AddBackground(frame, color)
    local hover = frame:CreateTexture(nil, "BACKGROUND");
    frame.hover = hover;
    hover:SetAllPoints();
    hover:SetColorTexture(unpack(color));
    return hover;
end


function fb:AddHoverBackground(frame, color)
    local hover = self:AddBackground(frame, color);
    hover:Hide();
    self:AddScript(frame, "OnEnter", function(self) self.hover:Show() end);
    self:AddScript(frame, "OnLeave", function(self) self.hover:Hide() end);
    -- frame:SetScript("OnEnter", function(self) self.hover:Show() end);
    -- frame:SetScript("OnLeave", function(self) self.hover:Hide() end);
end

--------------------

function fb:EnableTooltip(frame)
    if frame.__tooltipEnabled then return end
    frame.__tooltipEnabled = true;
	frame:EnableMouse();
	fb:AddScript(frame, "OnEnter", function(self) 
        if type(self.ShowTooltip) == "function" then
            self:ShowTooltip();
        elseif self.tooltip then
            GameTooltip:SetOwner(self);
            --GameTooltip:SetText("-");
            GameTooltip:AddLine(self.tooltip, nil, nil, nil, true)
            GameTooltip:SetWidth(200);
            GameTooltip:Show();
        end
	end);
	fb:AddScript(frame, "OnLeave", function() 
		GameTooltip:Hide();
	end);
end


function fb:SetTooltip(frame, arg)
    if arg then
        local a_type = type(arg);
        if a_type == "function" then
            frame.ShowTooltip = arg;
        elseif a_type == "string" then
            frame.tooltip = arg;
        end
    else
        frame.ShowTooltip = nil;
        frame.tooltip = nil;
    end
    self:EnableTooltip(frame);
end